@@ -33,6 +33,9 @@
 #include <ofono/log.h>
 #include <ofono/modem.h>
 #include <ofono/ussd.h>
+
+#include <drivers/atmodem/atutil.h>
+
 #include "util.h"
 #include "smsutil.h"
 #include "vendor.h"
@@ -40,8 +43,6 @@
 #include "gatchat.h"
 #include "gatresult.h"
 
-#include "atmodem.h"
-
 static const char *cusd_prefix[] = { "+CUSD:", NULL };
 static const char *none_prefix[] = { NULL };
 static const char *cscs_prefix[] = { "+CSCS:", NULL };
@@ -118,8 +119,10 @@ static void cusd_parse(GAtResult *result, struct ofono_ussd *ussd)
 	if (!g_at_result_iter_next_number(&iter, &status))
 		return;
 
-	if (!g_at_result_iter_next_string(&iter, &content))
+	if (!g_at_result_iter_next_string(&iter, &content)) {
+		dcs = 0;
 		goto out;
+	}
 
 	if (!g_at_result_iter_next_number(&iter, &dcs))
 		dcs = 0;
@@ -328,19 +331,10 @@ static void at_ussd_remove(struct ofono_ussd *ussd)
 }
 
 static const struct ofono_ussd_driver driver = {
-	.name		= "atmodem",
 	.probe		= at_ussd_probe,
 	.remove		= at_ussd_remove,
 	.request	= at_ussd_request,
 	.cancel		= at_ussd_cancel
 };
 
-void at_ussd_init(void)
-{
-	ofono_ussd_driver_register(&driver);
-}
-
-void at_ussd_exit(void)
-{
-	ofono_ussd_driver_unregister(&driver);
-}
+OFONO_ATOM_DRIVER_BUILTIN(ussd, atmodem, &driver)
