@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2009-2010 Nokia Corporation and/or its subsidiary(-ies).
+ *  Copyright (C) 2009-2010  Nokia Corporation and/or its subsidiary(-ies).
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -23,12 +23,12 @@
 #include <config.h>
 #endif
 
-#define _GNU_SOURCE
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 #include <errno.h>
 
+#include <ell/ell.h>
 #include <glib.h>
 
 #include <gisi/message.h>
@@ -39,7 +39,6 @@
 #include <ofono/modem.h>
 #include <ofono/netreg.h>
 
-#include "isimodem.h"
 #include "isiutil.h"
 #include "network.h"
 #include "debug.h"
@@ -75,12 +74,15 @@ struct network_time {
 
 struct netreg_data {
 	GIsiClient *client;
-	GIsiClient *primary;
-	GIsiClient *secondary;
+	GIsiClient *pn_network;
+	GIsiClient *pn_modem_network;
 	struct reg_info reg;
 	struct gsm_info gsm;
 	struct rat_info rat;
+	GIsiVersion version;
 	char nitz_name[OFONO_MAX_OPERATOR_NAME_LENGTH + 1];
+	char mcc[OFONO_MAX_MCC_LENGTH + 1];
+	char mnc[OFONO_MAX_MNC_LENGTH + 1];
 };
 
 static inline guint8 *mccmnc_to_bcd(const char *mcc, const char *mnc,
@@ -194,6 +196,12 @@ static void reg_status_ind_cb(const GIsiMessage *msg, void *data)
 	struct netreg_data *nd = ofono_netreg_get_data(netreg);
 	GIsiSubBlockIter iter;
 
+	uint8_t len = 0;
+	uint8_t type = 0;
+	char *abbrev = NULL;
+	char *an = NULL;
+	char *fn = NULL;
+
 	if (netreg == NULL || nd == NULL)
 		return;
 
@@ -209,17 +217,67 @@ static void reg_status_ind_cb(const GIsiMessage *msg, void *data)
 		case NET_REG_INFO_COMMON:
 
 			if (!parse_common_info(&iter, &nd->reg))
-				return;
+				break;
+
+			/* Parse abbreviated operator name */
+			if (!g_isi_sb_iter_get_byte(&iter, &len, 7))
+				break;
+
+			len *= 2;
+
+			g_isi_sb_iter_get_alpha_tag(&iter, &abbrev, len, 8);
 			break;
 
 		case NET_GSM_REG_INFO:
 
 			if (!parse_gsm_info(&iter, &nd->gsm))
-				return;
+				break;
+
+			/* Parse mcc and mnc */
+			g_isi_sb_iter_get_oper_code(&iter, nd->mcc,
+							   nd->mnc, 8);
+			break;
+
+		case NET_REG_STATUS_IND_OPER_NAME:
+
+			if (!g_isi_sb_iter_get_byte(&iter, &type, 2))
+				break;
+
+			if (!g_isi_sb_iter_get_byte(&iter, &len, 5))
+				break;
+
+			/* Name is UCS-2 encoded */
+			len *= 2;
+
+			switch (type) {
+			case 2:
+				/*
+				 * Alternate full name, which is similar to
+				 * long name, but may have less whitespace
+				 */
+				g_isi_sb_iter_get_alpha_tag(&iter, &an, len, 6);
+				break;
+			case 3:
+				/* Full name */
+				g_isi_sb_iter_get_alpha_tag(&iter, &fn, len, 6);
+				break;
+			}
+
 			break;
 		}
 	}
 
+	if (fn && fn[0] != '\0')
+		l_strlcpy(nd->nitz_name, fn, sizeof(nd->nitz_name));
+	else if (an && an[0] != '\0')
+		l_strlcpy(nd->nitz_name, an, sizeof(nd->nitz_name));
+	else if (abbrev && abbrev[0] != '\0')
+		l_strlcpy(nd->nitz_name, abbrev, sizeof(nd->nitz_name));
+
+	g_free(an);
+	g_free(fn);
+	g_free(abbrev);
+
 	ofono_netreg_status_notify(netreg, isi_status_to_at_status(&nd->reg),
 					nd->gsm.lac, nd->gsm.ci,
 					isi_to_at_tech(&nd->rat, &nd->gsm));
@@ -340,7 +398,8 @@ static void rat_resp_cb(const GIsiMessage *msg, void *data)
 			goto error;
 	}
 
-	if (g_isi_client_resource(nd->client) == PN_MODEM_NETWORK)
+	if (g_isi_client_resource(nd->client) == PN_MODEM_NETWORK ||
+			nd->version.major < 14)
 		req[0] = NET_MODEM_REG_STATUS_GET_REQ;
 
 	if (g_isi_client_send(nd->client, req, sizeof(req),
@@ -382,74 +441,6 @@ error:
 	g_free(cbd);
 }
 
-static void name_get_resp_cb(const GIsiMessage *msg, void *data)
-{
-	struct isi_cb_data *cbd = data;
-	ofono_netreg_operator_cb_t cb = cbd->cb;
-	struct ofono_network_operator op;
-
-	GIsiSubBlockIter iter;
-	uint8_t len = 0;
-	char *tag = NULL;
-
-	memset(&op, 0, sizeof(struct ofono_network_operator));
-
-	if (!check_response_status(msg, NET_OPER_NAME_READ_RESP))
-		goto error;
-
-	for (g_isi_sb_iter_init(&iter, msg, 6);
-			g_isi_sb_iter_is_valid(&iter);
-			g_isi_sb_iter_next(&iter)) {
-
-		switch (g_isi_sb_iter_get_id(&iter)) {
-		case NET_GSM_OPERATOR_INFO:
-
-			if (!g_isi_sb_iter_get_oper_code(&iter, op.mcc,
-								op.mnc, 2))
-				goto error;
-			break;
-
-		case NET_OPER_NAME_INFO:
-
-			if (!g_isi_sb_iter_get_byte(&iter, &len, 3))
-				goto error;
-
-			/* Name is UCS-2 encoded */
-			len *= 2;
-
-			if (!g_isi_sb_iter_get_alpha_tag(&iter, &tag, len, 4))
-				goto error;
-
-			strncpy(op.name, tag, OFONO_MAX_OPERATOR_NAME_LENGTH);
-			op.name[OFONO_MAX_OPERATOR_NAME_LENGTH] = '\0';
-			g_free(tag);
-			break;
-		}
-	}
-
-	CALLBACK_WITH_SUCCESS(cb, &op, cbd->data);
-	return;
-
-error:
-	CALLBACK_WITH_FAILURE(cb, NULL, cbd->data);
-}
-
-static gboolean send_name_read_req(GIsiClient *client, void *cbd,
-					GDestroyNotify destroy)
-{
-	const uint8_t msg[] = {
-		NET_OPER_NAME_READ_REQ,
-		NET_HARDCODED_LATIN_OPER_NAME,
-		OFONO_MAX_OPERATOR_NAME_LENGTH,
-		0x00, 0x00,	/* Index not used */
-		0x00,		/* Filler */
-		0x00,		/* No sub-blocks */
-	};
-
-	return g_isi_client_send(client, msg, sizeof(msg), name_get_resp_cb,
-					cbd, destroy);
-}
-
 static void cell_info_resp_cb(const GIsiMessage *msg, void *data)
 {
 	struct isi_cb_data *cbd = data;
@@ -499,15 +490,124 @@ error:
 	CALLBACK_WITH_FAILURE(cb, NULL, cbd->data);
 }
 
-static gboolean send_cell_info_req(GIsiClient *client, void *cbd,
-					GDestroyNotify destroy)
+static void create_cell_info_get_req(struct ofono_netreg *netreg,
+					ofono_netreg_operator_cb_t cb,
+					void *data)
 {
+	struct netreg_data *nd = ofono_netreg_get_data(netreg);
+	struct isi_cb_data *cbd = isi_cb_data_new(netreg, cb, data);
+
 	const uint8_t msg[] = {
 		NET_CELL_INFO_GET_REQ,
 	};
 
-	return g_isi_client_send(client, msg, sizeof(msg), cell_info_resp_cb,
-					cbd, destroy);
+	if (cbd == NULL || nd == NULL)
+		goto error;
+
+	if (g_isi_client_send(nd->client, msg, sizeof(msg), cell_info_resp_cb,
+				cbd, g_free))
+		return;
+
+error:
+	CALLBACK_WITH_FAILURE(cb, NULL, data);
+	g_free(cbd);
+}
+
+static void name_get_resp_cb(const GIsiMessage *msg, void *data)
+{
+	struct isi_cb_data *cbd = data;
+	ofono_netreg_operator_cb_t cb = cbd->cb;
+	struct ofono_netreg *netreg = cbd->user;
+	struct netreg_data *nd = ofono_netreg_get_data(netreg);
+	struct ofono_network_operator op;
+
+	GIsiSubBlockIter iter;
+	uint8_t len = 0;
+	char *tag = NULL;
+
+	memset(&op, 0, sizeof(struct ofono_network_operator));
+
+	if (!check_response_status(msg, NET_OLD_OPER_NAME_READ_RESP) &&
+			!check_response_status(msg, NET_OPER_NAME_READ_RESP))
+		goto alt_method;
+
+	for (g_isi_sb_iter_init(&iter, msg, 6);
+			g_isi_sb_iter_is_valid(&iter);
+			g_isi_sb_iter_next(&iter)) {
+
+		switch (g_isi_sb_iter_get_id(&iter)) {
+		case NET_GSM_OPERATOR_INFO:
+
+			if (!g_isi_sb_iter_get_oper_code(&iter, op.mcc,
+								op.mnc, 2))
+				goto error;
+			break;
+
+		case NET_OPER_NAME_INFO:
+
+			if (!g_isi_sb_iter_get_byte(&iter, &len, 3))
+				goto error;
+
+			/* Name is UCS-2 encoded */
+			len *= 2;
+
+			if (!g_isi_sb_iter_get_alpha_tag(&iter, &tag, len, 4))
+				goto error;
+
+			strncpy(op.name, tag, OFONO_MAX_OPERATOR_NAME_LENGTH);
+			op.name[OFONO_MAX_OPERATOR_NAME_LENGTH] = '\0';
+			g_free(tag);
+			break;
+		}
+	}
+
+success:
+	CALLBACK_WITH_SUCCESS(cb, &op, cbd->data);
+	return;
+
+alt_method:
+	if (nd->nitz_name[0] == '\0' || nd->mcc[0] == '\0' ||
+					nd->mnc[0] == '\0')
+		goto error;
+
+	l_strlcpy(op.name, nd->nitz_name, sizeof(op.name));
+	l_strlcpy(op.mcc, nd->mcc, sizeof(op.mcc));
+	l_strlcpy(op.mnc, nd->mnc, sizeof(op.mnc));
+	goto success;
+
+error:
+	CALLBACK_WITH_FAILURE(cb, NULL, cbd->data);
+}
+
+static void create_name_get_req(struct ofono_netreg *netreg,
+					ofono_netreg_operator_cb_t cb,
+					void *data)
+{
+	struct netreg_data *nd = ofono_netreg_get_data(netreg);
+	struct isi_cb_data *cbd = isi_cb_data_new(netreg, cb, data);
+
+	uint8_t msg[] = {
+		NET_OPER_NAME_READ_REQ,
+		NET_HARDCODED_LATIN_OPER_NAME,
+		OFONO_MAX_OPERATOR_NAME_LENGTH,
+		0x00, 0x00,	/* Index not used */
+		0x00,		/* Filler */
+		0x00,		/* No sub-blocks */
+	};
+
+	if (cbd == NULL || nd == NULL)
+		goto error;
+
+	if (nd->version.major < 14)
+		msg[0] = NET_OLD_OPER_NAME_READ_REQ;
+
+	if (g_isi_client_send(nd->client, msg, sizeof(msg), name_get_resp_cb,
+				cbd, g_free))
+		return;
+
+error:
+	CALLBACK_WITH_FAILURE(cb, NULL, data);
+	g_free(cbd);
 }
 
 static void isi_current_operator(struct ofono_netreg *netreg,
@@ -515,27 +615,11 @@ static void isi_current_operator(struct ofono_netreg *netreg,
 					void *data)
 {
 	struct netreg_data *nd = ofono_netreg_get_data(netreg);
-	struct isi_cb_data *cbd = isi_cb_data_new(netreg, cb, data);
 
-	if (cbd == NULL || nd == NULL)
-		goto error;
-
-	/*
-	 * PN_MODEM_NETWORK provides no operator name info, so getting
-	 * the current operator information relies on cell info
-	 * instead.
-	 */
-	if (g_isi_client_resource(nd->client) == PN_MODEM_NETWORK) {
-		if (send_cell_info_req(nd->client, cbd, g_free))
-			return;
-	} else {
-		if (send_name_read_req(nd->client, cbd, g_free))
-			return;
-	}
-
-error:
-	CALLBACK_WITH_FAILURE(cb, NULL, data);
-	g_free(cbd);
+	if (g_isi_client_resource(nd->client) == PN_MODEM_NETWORK)
+		create_cell_info_get_req(netreg, cb, data);
+	else
+		create_name_get_req(netreg, cb, data);
 }
 
 static void available_resp_cb(const GIsiMessage *msg, void *data)
@@ -655,7 +739,8 @@ static void isi_list_operators(struct ofono_netreg *netreg,
 	if (cbd == NULL || nd == NULL)
 		goto error;
 
-	if (g_isi_client_resource(nd->client) == PN_MODEM_NETWORK)
+	if (g_isi_client_resource(nd->client) == PN_MODEM_NETWORK ||
+			nd->version.major < 14)
 		msg[0] = NET_MODEM_AVAILABLE_GET_REQ;
 
 	if (g_isi_client_send_with_timeout(nd->client, msg, sizeof(msg),
@@ -802,10 +887,12 @@ static gboolean parse_nettime(GIsiSubBlockIter *iter,
 	info->min = time->min != NET_INVALID_TIME ? time->min : -1;
 	info->sec = time->sec != NET_INVALID_TIME ? time->sec : -1;
 
-	/* Most significant bit set indicates negative offset. The
+	/*
+	 * Most significant bit set indicates negative offset. The
 	 * second most significant bit is 'reserved'. The value is the
 	 * offset from UTCin a count of 15min intervals, possibly
-	 * including the current DST adjustment. */
+	 * including the current DST adjustment.
+	 */
 	info->utcoff = (time->utc & 0x3F) * 15 * 60;
 	if (time->utc & 0x80)
 		info->utcoff *= -1;
@@ -929,58 +1016,176 @@ error:
 	g_free(cbd);
 }
 
-static void common_client_setup(struct ofono_netreg *netreg,
-				struct netreg_data *nd)
+static void cs_access_config_resp_cb(const GIsiMessage *msg, void *data)
 {
-	g_isi_client_ind_subscribe(nd->client, NET_RSSI_IND, rssi_ind_cb,
-					netreg);
-	g_isi_client_ind_subscribe(nd->client, NET_NITZ_NAME_IND, name_ind_cb,
-					netreg);
-	g_isi_client_ind_subscribe(nd->client, NET_RAT_IND, rat_ind_cb, netreg);
-	g_isi_client_ind_subscribe(nd->client, NET_TIME_IND, time_ind_cb,
-					netreg);
-	g_isi_client_ind_subscribe(nd->client, NET_MODEM_REG_STATUS_IND,
-					reg_status_ind_cb, netreg);
-	g_isi_client_ind_subscribe(nd->client, NET_REG_STATUS_IND,
-					reg_status_ind_cb, netreg);
+	GIsiSubBlockIter iter;
+
+	DBG("");
+
+	if (g_isi_msg_id(msg) != NET_NW_ACCESS_CONF_RESP)
+		return;
+
+	/*
+	 * TODO: Check that roaming and registration
+	 * are now enabled.
+	 */
+
+	for (g_isi_sb_iter_init(&iter, msg, 2);
+			g_isi_sb_iter_is_valid(&iter);
+			g_isi_sb_iter_next(&iter)) {
+		uint8_t id = g_isi_sb_iter_get_id(&iter);
+		uint8_t mode;
+
+		DBG("SB=%02X", id);
+
+		switch (id) {
+		case NET_REGISTRATION_CONF_INFO:
+		case NET_REGISTRATION_CONF1_INFO:
+			g_isi_sb_iter_get_byte(&iter, &mode, 2);
+			DBG("Reg %X", mode);
+			break;
+
+		case NET_ROAMING_CONF_INFO:
+		case NET_ROAMING_CONF1_INFO:
+			g_isi_sb_iter_get_byte(&iter, &mode, 2);
+			DBG("Roam %X", mode);
+			break;
+
+		default:
+			DBG("Unknown subblock");
+		}
+	}
+}
+
+static void enable_registration(struct ofono_netreg *netreg)
+{
+	struct netreg_data *nd = ofono_netreg_get_data(netreg);
+	const uint8_t req[] = {
+		NET_NW_ACCESS_CONF_REQ, 0, 2,
+		/* Subblock 1 */
+		0x59, 4, 1, 0,
+		/* Subblock 2 */
+		0x5A, 4, 1, 0,
+	};
+
+	DBG("");
+	g_isi_client_send(nd->client, req, sizeof(req),
+				cs_access_config_resp_cb, netreg, NULL);
+}
+
+static void activate_cs_and_enable_registration(struct ofono_netreg *netreg)
+{
+	DBG("not implemented");
+}
+
+static void cs_state_resp_cb(const GIsiMessage *msg, void *data)
+{
+	struct ofono_netreg *netreg = data;
+	uint8_t code;
+
+	DBG("");
+
+	if (g_isi_msg_id(msg) != NET_CS_STATE_RESP)
+		return;
+
+	if (!g_isi_msg_data_get_byte(msg, 0, &code))
+		return;
+
+	if (code != NET_CAUSE_OK) {
+		DBG("Failed with cause=%X", code);
+		return;
+	}
+
+	if (!g_isi_msg_data_get_byte(msg, 1, &code))
+		return;
+
+	DBG("CS STATE=%X", code);
+
+	if (code == NET_CS_INACTIVE)
+		activate_cs_and_enable_registration(netreg);
+	else
+		enable_registration(netreg);
+}
+
+static void subscribe_indications(GIsiClient *cl, void *data)
+{
+	g_isi_client_ind_subscribe(cl, NET_RSSI_IND, rssi_ind_cb, data);
+	g_isi_client_ind_subscribe(cl, NET_NITZ_NAME_IND, name_ind_cb, data);
+	g_isi_client_ind_subscribe(cl, NET_RAT_IND, rat_ind_cb, data);
+	g_isi_client_ind_subscribe(cl, NET_TIME_IND, time_ind_cb, data);
+	g_isi_client_ind_subscribe(cl, NET_REG_STATUS_IND, reg_status_ind_cb,
+					data);
+	g_isi_client_ind_subscribe(cl, NET_MODEM_REG_STATUS_IND,reg_status_ind_cb,
+					data);
+}
+
+static void pn_network_reachable_cb(const GIsiMessage *msg, void *data)
+{
+	struct ofono_netreg *netreg = data;
+	struct netreg_data *nd = ofono_netreg_get_data(netreg);
+
+	if (g_isi_msg_error(msg) < 0) {
+		DBG("PN_NETWORK not reachable, removing client");
+		g_isi_client_destroy(nd->pn_network);
+		nd->pn_network = NULL;
+
+		if (nd->pn_modem_network == NULL)
+			ofono_netreg_remove(netreg);
+
+		return;
+	}
+
+	ISI_RESOURCE_DBG(msg);
+
+	if (nd == NULL || nd->client != NULL)
+		return;
+
+	nd->client = nd->pn_network;
+
+	nd->version.major = g_isi_msg_version_major(msg);
+	nd->version.minor = g_isi_msg_version_minor(msg);
+
+	subscribe_indications(nd->client, netreg);
 
 	ofono_netreg_register(netreg);
 }
 
-static void primary_reachable_cb(const GIsiMessage *msg, void *data)
+static void pn_modem_network_reachable_cb(const GIsiMessage *msg, void *data)
 {
 	struct ofono_netreg *netreg = data;
 	struct netreg_data *nd = ofono_netreg_get_data(netreg);
 
-	if (g_isi_msg_error(msg) < 0)
-		return;
+	const uint8_t req[] = {
+		NET_CS_STATE_REQ,
+	};
 
-	ISI_VERSION_DBG(msg);
+
+	if (g_isi_msg_error(msg) < 0) {
+		DBG("PN_MODEM_NETWORK not reachable, removing client");
+		g_isi_client_destroy(nd->pn_modem_network);
+		nd->pn_modem_network = NULL;
+
+		if (nd->pn_network == NULL)
+			ofono_netreg_remove(netreg);
+		return;
+	}
+
+	ISI_RESOURCE_DBG(msg);
 
 	if (nd == NULL || nd->client != NULL)
 		return;
 
-	nd->client = nd->primary;
+	nd->client = nd->pn_modem_network;
 
-	common_client_setup(netreg, nd);
-}
+	nd->version.major = g_isi_msg_version_major(msg);
+	nd->version.minor = g_isi_msg_version_minor(msg);
 
-static void secondary_reachable_cb(const GIsiMessage *msg, void *data)
-{
-	struct ofono_netreg *netreg = data;
-	struct netreg_data *nd = ofono_netreg_get_data(netreg);
+	subscribe_indications(nd->client, netreg);
 
-	if (g_isi_msg_error(msg) < 0)
-		return;
+	ofono_netreg_register(netreg);
 
-	ISI_VERSION_DBG(msg);
-
-	if (nd == NULL || nd->client != NULL)
-		return;
-
-	nd->client = nd->secondary;
-
-	common_client_setup(netreg, nd);
+	g_isi_client_send(nd->client, req, sizeof(req), cs_state_resp_cb,
+				netreg, NULL);
 }
 
 static int isi_netreg_probe(struct ofono_netreg *netreg, unsigned int vendor,
@@ -993,28 +1198,27 @@ static int isi_netreg_probe(struct ofono_netreg *netreg, unsigned int vendor,
 	if (nd == NULL)
 		return -ENOMEM;
 
-	nd->primary = g_isi_client_create(modem, PN_MODEM_NETWORK);
-	if (nd->primary == NULL)
-		goto nomem;
+	nd->pn_network = g_isi_client_create(modem, PN_NETWORK);
+	if (nd->pn_network == NULL) {
+		g_free(nd);
+		return -ENOMEM;
+	}
 
-	nd->secondary = g_isi_client_create(modem, PN_NETWORK);
-	if (nd->secondary == NULL)
-		goto nomem;
+	nd->pn_modem_network = g_isi_client_create(modem, PN_MODEM_NETWORK);
+	if (nd->pn_modem_network == NULL) {
+		g_isi_client_destroy(nd->pn_network);
+		g_free(nd);
+		return -ENOMEM;
+	}
 
 	ofono_netreg_set_data(netreg, nd);
 
-	g_isi_client_verify(nd->primary, primary_reachable_cb, netreg, NULL);
-	g_isi_client_verify(nd->secondary, secondary_reachable_cb, netreg,
-				NULL);
+	g_isi_client_verify(nd->pn_network, pn_network_reachable_cb,
+				netreg, NULL);
+	g_isi_client_verify(nd->pn_modem_network, pn_modem_network_reachable_cb,
+				netreg, NULL);
 
 	return 0;
-
-nomem:
-	g_isi_client_destroy(nd->primary);
-	g_isi_client_destroy(nd->secondary);
-
-	g_free(nd);
-	return -ENOMEM;
 }
 
 static void isi_netreg_remove(struct ofono_netreg *netreg)
@@ -1026,13 +1230,12 @@ static void isi_netreg_remove(struct ofono_netreg *netreg)
 	if (data == NULL)
 		return;
 
-	g_isi_client_destroy(data->primary);
-	g_isi_client_destroy(data->secondary);
+	g_isi_client_destroy(data->pn_modem_network);
+	g_isi_client_destroy(data->pn_network);
 	g_free(data);
 }
 
-static struct ofono_netreg_driver driver = {
-	.name			= "isimodem",
+static const struct ofono_netreg_driver driver = {
 	.probe			= isi_netreg_probe,
 	.remove			= isi_netreg_remove,
 	.registration_status	= isi_registration_status,
@@ -1043,12 +1246,4 @@ static struct ofono_netreg_driver driver = {
 	.strength		= isi_strength,
 };
 
-void isi_netreg_init(void)
-{
-	ofono_netreg_driver_register(&driver);
-}
-
-void isi_netreg_exit(void)
-{
-	ofono_netreg_driver_unregister(&driver);
-}
+OFONO_ATOM_DRIVER_BUILTIN(netreg, isimodem, &driver)
