@@ -54,7 +54,6 @@
 #include <ofono/gprs-context.h>
 #include <ofono/audio-settings.h>
 
-#include "drivers/isimodem/isimodem.h"
 #include "drivers/isimodem/isiutil.h"
 #include "drivers/isimodem/infoserver.h"
 #include "drivers/isimodem/mtc.h"
@@ -238,10 +237,10 @@ static void mtc_reachable_cb(const GIsiMessage *msg, void *data)
 	struct ofono_modem *modem = data;
 	struct isi_data *isi = ofono_modem_get_data(modem);
 
-	if (!g_isi_msg_error(msg) < 0)
+	if (g_isi_msg_error(msg) < 0)
 		return;
 
-	ISI_VERSION_DBG(msg);
+	ISI_RESOURCE_DBG(msg);
 
 	g_isi_client_ind_subscribe(isi->client, MTC_STATE_INFO_IND,
 					mtc_state_ind_cb, modem);
