@@ -27,11 +27,13 @@
 #include <stdlib.h>
 
 #include <glib.h>
+#include <ell/ell.h>
 
 #include <ofono/types.h>
 #include "simutil.h"
 #include "util.h"
 #include "smsutil.h"
+#include "missing.h"
 
 struct sim_eons {
 	struct sim_eons_operator_info *pnn_list;
@@ -764,6 +766,20 @@ unsigned char *comprehension_tlv_builder_get_data(
 	return tlv + tag_size + len_size;
 }
 
+bool validate_utf8_tlv(const unsigned char *tlv)
+{
+	int len = tlv[1];
+
+	if (len == 0)
+		return FALSE;
+
+	/* support both null-terminated and non null-terminated TLV value */
+	if (tlv[len + 1] == '\0')
+		len -= 1;
+
+	return l_utf8_validate((const char *)tlv + 2, len, NULL);
+}
+
 static char *sim_network_name_parse(const unsigned char *buffer, int length,
 					gboolean *add_ci)
 {
@@ -793,14 +809,14 @@ static char *sim_network_name_parse(const unsigned char *buffer, int length,
 		spare_bits = dcs & 0x07;
 		num_char = (length * 8 - spare_bits) / 7;
 
-		unpacked_buf = unpack_7bit(buffer, length, 0, FALSE,
+		unpacked_buf = unpack_7bit(buffer, length, 0, false,
 						num_char, &written, 0);
 		if (unpacked_buf == NULL)
 			break;
 
 		ret = convert_gsm_to_utf8(unpacked_buf, written, NULL, NULL, 0);
 
-		g_free(unpacked_buf);
+		l_free(unpacked_buf);
 
 		break;
 	case 0x10:
@@ -815,9 +831,7 @@ static char *sim_network_name_parse(const unsigned char *buffer, int length,
 			if (buffer[i] == 0xff && buffer[i + 1] == 0xff)
 				break;
 
-		ret = g_convert((const char *) buffer, length,
-					"UTF-8//TRANSLIT", "UCS-2BE",
-					NULL, NULL, NULL);
+		ret = l_utf8_from_ucs2be(buffer, length);
 		break;
 	}
 
@@ -989,9 +1003,9 @@ static void pnn_operator_free(struct sim_eons_operator_info *oper)
 	if (oper == NULL)
 		return;
 
-	g_free(oper->info);
-	g_free(oper->shortname);
-	g_free(oper->longname);
+	l_free(oper->info);
+	l_free(oper->shortname);
+	l_free(oper->longname);
 }
 
 struct sim_eons *sim_eons_new(int pnn_records)
@@ -1411,7 +1425,7 @@ gboolean sim_parse_3g_get_response(const unsigned char *data, int len,
 	if (tlv[1] != 0x21)
 		return FALSE;
 
-	switch (tlv[0] & 0x3) {
+	switch (tlv[0] & 0x7) {
 	case 1:	/* Transparent */
 		str = 0x00;
 		break;
@@ -1551,7 +1565,7 @@ gboolean sim_cphs_is_active(unsigned char *cphs, enum sim_cphs_service index)
 
 void sim_app_record_free(struct sim_app_record *app)
 {
-	g_free(app->label);
+	l_free(app->label);
 	g_free(app);
 }
 
@@ -1575,6 +1589,7 @@ GSList *sim_parse_app_template_entries(const unsigned char *buffer, int len)
 			goto error;
 
 		memcpy(app.aid, aid, app.aid_len);
+		memset(app.aid + app.aid_len, 0xff, 16 - app.aid_len);
 
 		app.type = (app.aid[5] << 8) | app.aid[6];
 
@@ -1593,7 +1608,7 @@ GSList *sim_parse_app_template_entries(const unsigned char *buffer, int len)
 		} else
 			app.label = NULL;
 
-		ret = g_slist_prepend(ret, g_memdup(&app, sizeof(app)));
+		ret = g_slist_prepend(ret, g_memdup2(&app, sizeof(app)));
 
 		len -= (dataobj - buffer) + dataobj_len;
 		buffer = dataobj + dataobj_len;
@@ -1606,7 +1621,7 @@ error:
 		GSList *t = ret;
 		struct sim_app_record *app = ret->data;
 
-		g_free(app->label);
+		l_free(app->label);
 		g_free(app);
 
 		ret = ret->next;
@@ -1659,67 +1674,135 @@ int sim_build_gsm_authenticate(unsigned char *buffer, int len,
 	return build_authenticate(buffer, rand, NULL);
 }
 
-gboolean sim_parse_umts_authenticate(const unsigned char *buffer,
-		int len, const unsigned char **res, const unsigned char **ck,
-		const unsigned char **ik, const unsigned char **auts,
-		const unsigned char **kc)
+gboolean sim_parse_umts_authenticate(const unsigned char *buffer, int len,
+		struct data_block *res, struct data_block *ck,
+		struct data_block *ik, struct data_block *auts,
+		struct data_block *sres, struct data_block *kc)
 {
-	if (len < 16 || !buffer)
+	const unsigned char *ptr = buffer;
+	const unsigned char *end = ptr + len;
+	unsigned int l;
+
+	if (!buffer || len < 2)
 		return FALSE;
 
-	switch (buffer[0]) {
+	memset(res, 0, sizeof(*res));
+	memset(ck, 0, sizeof(*ck));
+	memset(ik, 0, sizeof(*ik));
+	memset(kc, 0, sizeof(*kc));
+	memset(auts, 0, sizeof(*auts));
+	memset(sres, 0, sizeof(*sres));
+
+	/*
+	 * TS 31.102
+	 * 7.1.2.1 GSM/3G security context
+	 */
+	switch (*ptr++) {
 	case 0xdb:
-		/* 'DB' + '08' + RES(16) + '10' + CK(32) + '10' + IK(32) = 43 */
-		if (len < 43)
-			goto umts_end;
+		/*
+		 * Response parameters/data, case 1, 3G security context,
+		 * command successful:
+		 *
+		 * "Successful 3G authentication" tag = 'DB'
+		 * 'DB' + L3 + RES(L3) + L4 + CK(L4) + L5 + IK(L5) + 8 + Kc(8)
+		 */
+		l = *ptr++; /* L3 */
+		if ((ptr + l) > end)
+			return FALSE;
 
-		/* success */
-		if (buffer[1] != 0x08)
-			goto umts_end;
+		res->data = ptr;
+		res->len = l;
+		ptr += l;
 
-		*res = buffer + 2;
+		if (ptr == end)
+			return FALSE;
 
-		if (buffer[10] != 0x10)
-			goto umts_end;
+		l = *ptr++; /* L4 */
+		if ((ptr + l) > end)
+			return FALSE;
 
-		*ck = buffer + 11;
+		ck->data = ptr;
+		ck->len = l;
+		ptr += l;
 
-		if (buffer[27] != 0x10)
-			goto umts_end;
+		if (ptr == end)
+			return FALSE;
 
-		*ik = buffer + 28;
+		l = *ptr++; /* L5 */
+		if ((ptr + l) > end)
+			return FALSE;
 
-		if (len >= 53 && kc) {
-			if (buffer[44] != 0x08)
-				goto umts_end;
+		ik->data = ptr;
+		ik->len = l;
+		ptr += l;
 
-			*kc = buffer + 45;
-		} else {
-			*kc = NULL;
+		if (ptr < end) {
+			l = *ptr++;
+			if (l != 8 || (ptr + l) != end)
+				return FALSE;
+
+			kc->data = ptr;
+			kc->len = l;
+			ptr += l;
 		}
 
-		*auts = NULL;
+		return TRUE;
 
-		break;
 	case 0xdc:
-		/* 'DC' + '0E' + AUTS(14) = 16 */
-		if (len < 16)
-			goto umts_end;
+		/*
+		 * Response parameters/data, case 2, 3G security context,
+		 * synchronisation failure:
+		 *
+		 * "Synchronisation failure" tag = 'DC'
+		 * 'DC' + L1 + AUTS(L1)
+		 */
+		l = *ptr++; /* L1 */
+		if ((ptr + l) > end)
+			return FALSE;
 
-		/* sync error */
-		if (buffer[1] != 0x0e)
-			goto umts_end;
+		auts->data = ptr;
+		auts->len = l;
+		ptr += l;
 
-		*auts = buffer + 2;
+		if (ptr != end)
+			return FALSE;
+
+		return TRUE;
+
+	case 0x04:
+		/*
+		 * Response parameters/data, case 3, GSM security context,
+		 * command successful:
+		 *
+		 * 4 + SRES(4) + 8 + Kc(8)
+		 */
+		l = 4; /* Already skipped this one */
+		if ((ptr + l) > end)
+			return FALSE;
+
+		sres->data = ptr;
+		sres->len = l;
+		ptr += l;
+
+		if (ptr == end)
+			return FALSE;
+
+		l = *ptr++; /* 8 */
+		if (l != 8 || (ptr + l) > end)
+			return FALSE;
+
+		kc->data = ptr;
+		kc->len = l;
+		ptr += l;
+
+		if (ptr != end)
+			return FALSE;
+
+		return TRUE;
 
-		break;
 	default:
-		goto umts_end;
+		break;
 	}
-
-	return TRUE;
-
-umts_end:
 	return FALSE;
 }
 
