@@ -2,7 +2,7 @@
  *
  *  oFono - Open Source Telephony
  *
- *  Copyright (C) 2008-2010  Intel Corporation. All rights reserved.
+ *  Copyright (C) 2008-2011  Intel Corporation. All rights reserved.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License version 2 as
@@ -23,13 +23,13 @@
 #include <config.h>
 #endif
 
-#define _GNU_SOURCE
 #include <stdint.h>
 #include <string.h>
 #include <errno.h>
 
 #include <glib.h>
 #include <gdbus.h>
+#include <ell/ell.h>
 
 #include "ofono.h"
 
@@ -38,6 +38,10 @@
 #include "stkutil.h"
 #include "stkagent.h"
 
+#ifndef DBUS_TIMEOUT_INFINITE
+#define DBUS_TIMEOUT_INFINITE ((int) 0x7fffffff)
+#endif
+
 enum allowed_error {
 	ALLOWED_ERROR_GO_BACK	= 0x1,
 	ALLOWED_ERROR_TERMINATE	= 0x2,
@@ -48,13 +52,16 @@ struct stk_agent {
 	char *path;				/* Agent Path */
 	char *bus;				/* Agent bus */
 	guint disconnect_watch;			/* DBus disconnect watch */
-	ofono_bool_t remove_on_terminate;
+	bool remove_on_terminate;
 	ofono_destroy_func removed_cb;
 	void *removed_data;
 	DBusMessage *msg;
 	DBusPendingCall *call;
 	void *user_cb;
 	void *user_data;
+	int min_length;
+	int max_length;
+	bool hidden_entry;
 	ofono_destroy_func user_destroy;
 
 	const struct stk_menu *request_selection_menu;
@@ -111,7 +118,7 @@ static void stk_agent_request_end(struct stk_agent *agent)
 	agent->user_cb = NULL;
 }
 
-ofono_bool_t stk_agent_matches(struct stk_agent *agent,
+bool stk_agent_matches(struct stk_agent *agent,
 				const char *path, const char *sender)
 {
 	return !strcmp(agent->path, path) && !strcmp(agent->bus, sender);
@@ -221,7 +228,7 @@ static void stk_agent_disconnect_cb(DBusConnection *conn, void *user_data)
 }
 
 struct stk_agent *stk_agent_new(const char *path, const char *sender,
-				ofono_bool_t remove_on_terminate)
+				bool remove_on_terminate)
 {
 	struct stk_agent *agent = g_try_new0(struct stk_agent, 1);
 	DBusConnection *conn = ofono_dbus_get_connection();
@@ -281,9 +288,9 @@ void append_menu_items_variant(DBusMessageIter *iter,
 done:								\
 	if (result == STK_AGENT_RESULT_TERMINATE &&		\
 			agent->remove_on_terminate)		\
-		remove_agent = TRUE;				\
+		remove_agent = true;				\
 	else							\
-		remove_agent = FALSE;				\
+		remove_agent = false;				\
 								\
 error:								\
 	stk_agent_request_end(agent);				\
@@ -300,12 +307,12 @@ static void request_selection_cb(DBusPendingCall *call, void *data)
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	unsigned char selection, i;
 	enum stk_agent_result result;
-	gboolean remove_agent;
+	bool remove_agent;
 
 	if (check_error(agent, reply,
 			ALLOWED_ERROR_GO_BACK | ALLOWED_ERROR_TERMINATE,
 			&result) == -EINVAL) {
-		remove_agent = TRUE;
+		remove_agent = true;
 		goto error;
 	}
 
@@ -318,7 +325,7 @@ static void request_selection_cb(DBusPendingCall *call, void *data)
 					DBUS_TYPE_BYTE, &selection,
 					DBUS_TYPE_INVALID) == FALSE) {
 		ofono_error("Can't parse the reply to RequestSelection()");
-		remove_agent = TRUE;
+		remove_agent = true;
 		goto error;
 	}
 
@@ -326,7 +333,7 @@ static void request_selection_cb(DBusPendingCall *call, void *data)
 
 	if (i != selection) {
 		ofono_error("Invalid item selected");
-		remove_agent = TRUE;
+		remove_agent = true;
 		goto error;
 	}
 
@@ -381,12 +388,12 @@ static void display_text_cb(DBusPendingCall *call, void *data)
 	stk_agent_display_text_cb cb = agent->user_cb;
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	enum stk_agent_result result;
-	gboolean remove_agent;
+	bool remove_agent;
 
 	if (check_error(agent, reply,
 			ALLOWED_ERROR_GO_BACK | ALLOWED_ERROR_TERMINATE |
 			ALLOWED_ERROR_BUSY, &result) == -EINVAL) {
-		remove_agent = TRUE;
+		remove_agent = true;
 		goto error;
 	}
 
@@ -397,7 +404,7 @@ static void display_text_cb(DBusPendingCall *call, void *data)
 
 	if (dbus_message_get_args(reply, NULL, DBUS_TYPE_INVALID) == FALSE) {
 		ofono_error("Can't parse the reply to DisplayText()");
-		remove_agent = TRUE;
+		remove_agent = true;
 		goto error;
 	}
 
@@ -408,7 +415,7 @@ static void display_text_cb(DBusPendingCall *call, void *data)
 
 int stk_agent_display_text(struct stk_agent *agent, const char *text,
 				const struct stk_icon_id *icon,
-				ofono_bool_t urgent,
+				bool urgent,
 				stk_agent_display_text_cb cb,
 				void *user_data, ofono_destroy_func destroy,
 				int timeout)
@@ -449,18 +456,18 @@ static void get_confirmation_cb(DBusPendingCall *call, void *data)
 	stk_agent_confirmation_cb cb = agent->user_cb;
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	enum stk_agent_result result;
-	gboolean remove_agent;
+	bool remove_agent;
 	dbus_bool_t confirm;
 
 	if (check_error(agent, reply,
 			ALLOWED_ERROR_GO_BACK | ALLOWED_ERROR_TERMINATE,
 			&result) == -EINVAL) {
-		remove_agent = TRUE;
+		remove_agent = true;
 		goto error;
 	}
 
 	if (result != STK_AGENT_RESULT_OK) {
-		cb(result, FALSE, agent->user_data);
+		cb(result, false, agent->user_data);
 		goto done;
 	}
 
@@ -468,7 +475,7 @@ static void get_confirmation_cb(DBusPendingCall *call, void *data)
 					DBUS_TYPE_BOOLEAN, &confirm,
 					DBUS_TYPE_INVALID) == FALSE) {
 		ofono_error("Can't parse the reply to GetConfirmation()");
-		remove_agent = TRUE;
+		remove_agent = true;
 		goto error;
 	}
 
@@ -518,13 +525,13 @@ static void get_digit_cb(DBusPendingCall *call, void *data)
 	stk_agent_string_cb cb = agent->user_cb;
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	enum stk_agent_result result;
-	gboolean remove_agent;
+	bool remove_agent;
 	char *digit;
 
 	if (check_error(agent, reply,
 			ALLOWED_ERROR_GO_BACK | ALLOWED_ERROR_TERMINATE,
 			&result) == -EINVAL) {
-		remove_agent = TRUE;
+		remove_agent = true;
 		goto error;
 	}
 
@@ -535,11 +542,21 @@ static void get_digit_cb(DBusPendingCall *call, void *data)
 
 	if (dbus_message_get_args(reply, NULL,
 					DBUS_TYPE_STRING, &digit,
-					DBUS_TYPE_INVALID) == FALSE ||
-			strlen(digit) != 1 ||
-			!valid_phone_number_format(digit)) {
+					DBUS_TYPE_INVALID) == FALSE) {
 		ofono_error("Can't parse the reply to GetDigit()");
-		remove_agent = TRUE;
+		remove_agent = true;
+		goto error;
+	}
+
+	if (strlen(digit) != 1 || !strspn(digit, "0123456789*#+")) {
+		ofono_error("Invalid character");
+		remove_agent = true;
+		goto error;
+	}
+
+	if (agent->hidden_entry && digit[0] == '+') {
+		ofono_error("The character + is not allowed in this mode");
+		remove_agent = true;
 		goto error;
 	}
 
@@ -574,6 +591,40 @@ int stk_agent_request_digit(struct stk_agent *agent, const char *text,
 	agent->user_cb = cb;
 	agent->user_data = user_data;
 	agent->user_destroy = destroy;
+	agent->hidden_entry = false;
+
+	dbus_pending_call_set_notify(agent->call, get_digit_cb, agent, NULL);
+
+	return 0;
+}
+
+int stk_agent_request_quick_digit(struct stk_agent *agent, const char *text,
+					const struct stk_icon_id *icon,
+					stk_agent_string_cb cb, void *user_data,
+					ofono_destroy_func destroy, int timeout)
+{
+	DBusConnection *conn = ofono_dbus_get_connection();
+
+	agent->msg = dbus_message_new_method_call(agent->bus, agent->path,
+							OFONO_SIM_APP_INTERFACE,
+							"RequestQuickDigit");
+	if (agent->msg == NULL)
+		return -ENOMEM;
+
+	dbus_message_append_args(agent->msg,
+					DBUS_TYPE_STRING, &text,
+					DBUS_TYPE_BYTE, &icon->id,
+					DBUS_TYPE_INVALID);
+
+	if (dbus_connection_send_with_reply(conn, agent->msg, &agent->call,
+						timeout) == FALSE ||
+			agent->call == NULL)
+		return -EIO;
+
+	agent->user_cb = cb;
+	agent->user_data = user_data;
+	agent->user_destroy = destroy;
+	agent->hidden_entry = true;
 
 	dbus_pending_call_set_notify(agent->call, get_digit_cb, agent, NULL);
 
@@ -586,13 +637,15 @@ static void get_key_cb(DBusPendingCall *call, void *data)
 	stk_agent_string_cb cb = agent->user_cb;
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	enum stk_agent_result result;
-	gboolean remove_agent;
+	bool remove_agent;
 	char *key;
+	int len;
+	wchar_t cp;
 
 	if (check_error(agent, reply,
 			ALLOWED_ERROR_GO_BACK | ALLOWED_ERROR_TERMINATE,
 			&result) == -EINVAL) {
-		remove_agent = TRUE;
+		remove_agent = true;
 		goto error;
 	}
 
@@ -603,10 +656,16 @@ static void get_key_cb(DBusPendingCall *call, void *data)
 
 	if (dbus_message_get_args(reply, NULL,
 					DBUS_TYPE_STRING, &key,
-					DBUS_TYPE_INVALID) == FALSE ||
-			g_utf8_strlen(key, 10) != 1) {
+					DBUS_TYPE_INVALID) == FALSE) {
 		ofono_error("Can't parse the reply to GetKey()");
-		remove_agent = TRUE;
+		remove_agent = true;
+		goto error;
+	}
+
+	len = strlen(key);
+	if (l_utf8_get_codepoint(key, len, &cp) != len) {
+		ofono_error("GetKey() return expected a single character");
+		remove_agent = true;
 		goto error;
 	}
 
@@ -617,7 +676,7 @@ static void get_key_cb(DBusPendingCall *call, void *data)
 
 int stk_agent_request_key(struct stk_agent *agent, const char *text,
 				const struct stk_icon_id *icon,
-				ofono_bool_t unicode_charset,
+				bool unicode_charset,
 				stk_agent_string_cb cb, void *user_data,
 				ofono_destroy_func destroy, int timeout)
 {
@@ -654,13 +713,14 @@ static void get_digits_cb(DBusPendingCall *call, void *data)
 	stk_agent_string_cb cb = agent->user_cb;
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	enum stk_agent_result result;
-	gboolean remove_agent;
+	bool remove_agent;
 	char *string;
+	int len, span;
 
 	if (check_error(agent, reply,
 			ALLOWED_ERROR_GO_BACK | ALLOWED_ERROR_TERMINATE,
 			&result) == -EINVAL) {
-		remove_agent = TRUE;
+		remove_agent = true;
 		goto error;
 	}
 
@@ -673,7 +733,26 @@ static void get_digits_cb(DBusPendingCall *call, void *data)
 					DBUS_TYPE_STRING, &string,
 					DBUS_TYPE_INVALID) == FALSE) {
 		ofono_error("Can't parse the reply to GetDigits()");
-		remove_agent = TRUE;
+		remove_agent = true;
+		goto error;
+	}
+
+	len = strlen(string);
+
+	if (len < agent->min_length || len > agent->max_length) {
+		ofono_error("Length not acceptable");
+		remove_agent = true;
+		goto error;
+	}
+
+	if (agent->hidden_entry)
+		span = strspn(string, "0123456789*#");
+	else
+		span = strspn(string, "0123456789*#+");
+
+	if (span != len) {
+		ofono_error("Invalid character found");
+		remove_agent = true;
 		goto error;
 	}
 
@@ -685,7 +764,7 @@ static void get_digits_cb(DBusPendingCall *call, void *data)
 int stk_agent_request_digits(struct stk_agent *agent, const char *text,
 				const struct stk_icon_id *icon,
 				const char *default_text,
-				int min, int max, ofono_bool_t hidden,
+				int min, int max, bool hidden,
 				stk_agent_string_cb cb, void *user_data,
 				ofono_destroy_func destroy, int timeout)
 {
@@ -720,6 +799,9 @@ int stk_agent_request_digits(struct stk_agent *agent, const char *text,
 	agent->user_cb = cb;
 	agent->user_data = user_data;
 	agent->user_destroy = destroy;
+	agent->min_length = min_val;
+	agent->max_length = max_val;
+	agent->hidden_entry = hidden_val;
 
 	dbus_pending_call_set_notify(agent->call, get_digits_cb, agent, NULL);
 
@@ -732,13 +814,14 @@ static void get_input_cb(DBusPendingCall *call, void *data)
 	stk_agent_string_cb cb = agent->user_cb;
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	enum stk_agent_result result;
-	gboolean remove_agent;
+	bool remove_agent;
 	char *string;
+	int len;
 
 	if (check_error(agent, reply,
 			ALLOWED_ERROR_GO_BACK | ALLOWED_ERROR_TERMINATE,
 			&result) == -EINVAL) {
-		remove_agent = TRUE;
+		remove_agent = true;
 		goto error;
 	}
 
@@ -751,7 +834,15 @@ static void get_input_cb(DBusPendingCall *call, void *data)
 					DBUS_TYPE_STRING, &string,
 					DBUS_TYPE_INVALID) == FALSE) {
 		ofono_error("Can't parse the reply to GetInput()");
-		remove_agent = TRUE;
+		remove_agent = true;
+		goto error;
+	}
+
+	len = l_utf8_strlen(string);
+
+	if (len < agent->min_length || len > agent->max_length) {
+		ofono_error("Length not acceptable");
+		remove_agent = true;
 		goto error;
 	}
 
@@ -763,8 +854,8 @@ static void get_input_cb(DBusPendingCall *call, void *data)
 int stk_agent_request_input(struct stk_agent *agent, const char *text,
 				const struct stk_icon_id *icon,
 				const char *default_text,
-				ofono_bool_t unicode_charset, int min, int max,
-				ofono_bool_t hidden, stk_agent_string_cb cb,
+				bool unicode_charset, int min, int max,
+				bool hidden, stk_agent_string_cb cb,
 				void *user_data, ofono_destroy_func destroy,
 				int timeout)
 {
@@ -799,6 +890,9 @@ int stk_agent_request_input(struct stk_agent *agent, const char *text,
 	agent->user_cb = cb;
 	agent->user_data = user_data;
 	agent->user_destroy = destroy;
+	agent->min_length = min_val;
+	agent->max_length = max_val;
+	agent->hidden_entry = hidden_val;
 
 	dbus_pending_call_set_notify(agent->call, get_input_cb, agent, NULL);
 
@@ -811,12 +905,12 @@ static void confirm_call_cb(DBusPendingCall *call, void *data)
 	stk_agent_confirmation_cb cb = agent->user_cb;
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	enum stk_agent_result result;
-	gboolean remove_agent;
+	bool remove_agent;
 	dbus_bool_t confirm;
 
 	if (check_error(agent, reply,
 			ALLOWED_ERROR_TERMINATE, &result) == -EINVAL) {
-		remove_agent = TRUE;
+		remove_agent = true;
 		goto error;
 	}
 
@@ -829,7 +923,7 @@ static void confirm_call_cb(DBusPendingCall *call, void *data)
 					DBUS_TYPE_BOOLEAN, &confirm,
 					DBUS_TYPE_INVALID) == FALSE) {
 		ofono_error("Can't parse the reply to ConfirmCallSetup()");
-		remove_agent = TRUE;
+		remove_agent = true;
 		goto error;
 	}
 
@@ -877,17 +971,17 @@ static void play_tone_cb(DBusPendingCall *call, void *data)
 	stk_agent_tone_cb cb = agent->user_cb;
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	enum stk_agent_result result;
-	gboolean remove_agent;
+	bool remove_agent;
 
 	if (check_error(agent, reply,
 			ALLOWED_ERROR_TERMINATE, &result) == -EINVAL) {
-		remove_agent = TRUE;
+		remove_agent = true;
 		goto error;
 	}
 
 	if (dbus_message_get_args(reply, NULL, DBUS_TYPE_INVALID) == FALSE) {
 		ofono_error("Can't parse the reply to PlayTone()");
-		remove_agent = TRUE;
+		remove_agent = true;
 		goto error;
 	}
 
@@ -898,7 +992,7 @@ static void play_tone_cb(DBusPendingCall *call, void *data)
 }
 
 int stk_agent_play_tone(struct stk_agent *agent, const char *text,
-			const struct stk_icon_id *icon, ofono_bool_t vibrate,
+			const struct stk_icon_id *icon, bool vibrate,
 			const char *tone, stk_agent_tone_cb cb, void *user_data,
 			ofono_destroy_func destroy, int timeout)
 {
@@ -932,7 +1026,7 @@ int stk_agent_play_tone(struct stk_agent *agent, const char *text,
 }
 
 int stk_agent_loop_tone(struct stk_agent *agent, const char *text,
-			const struct stk_icon_id *icon, ofono_bool_t vibrate,
+			const struct stk_icon_id *icon, bool vibrate,
 			const char *tone, stk_agent_tone_cb cb, void *user_data,
 			ofono_destroy_func destroy, int timeout)
 {
@@ -970,16 +1064,16 @@ static void action_info_cb(DBusPendingCall *call, void *data)
 	struct stk_agent *agent = data;
 	DBusMessage *reply = dbus_pending_call_steal_reply(call);
 	enum stk_agent_result result;
-	gboolean remove_agent;
+	bool remove_agent;
 
 	if (check_error(agent, reply, 0, &result) == -EINVAL) {
-		remove_agent = TRUE;
+		remove_agent = true;
 		goto error;
 	}
 
 	if (dbus_message_get_args(reply, NULL, DBUS_TYPE_INVALID) == FALSE) {
 		ofono_error("Can't parse the reply to DisplayActionInfo()");
-		remove_agent = TRUE;
+		remove_agent = true;
 		goto error;
 	}
 
@@ -1005,7 +1099,7 @@ int stk_agent_display_action_info(struct stk_agent *agent, const char *text,
 					DBUS_TYPE_INVALID);
 
 	if (dbus_connection_send_with_reply(conn, agent->msg, &agent->call,
-						0) == FALSE ||
+					DBUS_TIMEOUT_INFINITE) == FALSE ||
 			agent->call == NULL)
 		return -EIO;
 
@@ -1024,7 +1118,7 @@ static void confirm_launch_browser_cb(DBusPendingCall *call, void *data)
 	dbus_bool_t confirm;
 
 	if (check_error(agent, reply, 0, &result) == -EINVAL) {
-		remove_agent = TRUE;
+		remove_agent = true;
 		cb(STK_AGENT_RESULT_TERMINATE, FALSE, agent->user_data);
 		goto error;
 	}
@@ -1038,7 +1132,7 @@ static void confirm_launch_browser_cb(DBusPendingCall *call, void *data)
 					DBUS_TYPE_BOOLEAN, &confirm,
 					DBUS_TYPE_INVALID) == FALSE) {
 		ofono_error("Can't parse the reply to ConfirmLaunchBrowser()");
-		remove_agent = TRUE;
+		remove_agent = true;
 		goto error;
 	}
 
@@ -1061,6 +1155,9 @@ int stk_agent_confirm_launch_browser(struct stk_agent *agent, const char *text,
 	if (agent->msg == NULL)
 		return -ENOMEM;
 
+	if (url == NULL)
+		url = "";
+
 	dbus_message_append_args(agent->msg,
 					DBUS_TYPE_STRING, &text,
 					DBUS_TYPE_BYTE, &icon_id,
@@ -1081,3 +1178,131 @@ int stk_agent_confirm_launch_browser(struct stk_agent *agent, const char *text,
 
 	return 0;
 }
+
+static void display_action_cb(DBusPendingCall *call, void *data)
+{
+	struct stk_agent *agent = data;
+	stk_agent_display_action_cb cb = agent->user_cb;
+	DBusMessage *reply = dbus_pending_call_steal_reply(call);
+	enum stk_agent_result result;
+	gboolean remove_agent;
+
+	if (check_error(agent, reply,
+			ALLOWED_ERROR_TERMINATE, &result) == -EINVAL) {
+		remove_agent = true;
+		goto error;
+	}
+
+	if (dbus_message_get_args(reply, NULL, DBUS_TYPE_INVALID) == FALSE) {
+		ofono_error("Can't parse the reply to DisplayAction()");
+		remove_agent = true;
+		goto error;
+	}
+
+	cb(result, agent->user_data);
+	goto done;
+
+	CALLBACK_END();
+}
+
+int stk_agent_display_action(struct stk_agent *agent,
+					const char *text,
+					const struct stk_icon_id *icon,
+					stk_agent_display_action_cb cb,
+					void *user_data,
+					ofono_destroy_func destroy)
+{
+	DBusConnection *conn = ofono_dbus_get_connection();
+
+	agent->msg = dbus_message_new_method_call(agent->bus, agent->path,
+						OFONO_SIM_APP_INTERFACE,
+						"DisplayAction");
+	if (agent->msg == NULL)
+		return -ENOMEM;
+
+	dbus_message_append_args(agent->msg,
+					DBUS_TYPE_STRING, &text,
+					DBUS_TYPE_BYTE, &icon->id,
+					DBUS_TYPE_INVALID);
+
+	if (dbus_connection_send_with_reply(conn, agent->msg, &agent->call,
+					DBUS_TIMEOUT_INFINITE) == FALSE ||
+			agent->call == NULL)
+		return -EIO;
+
+	agent->user_cb = cb;
+	agent->user_data = user_data;
+	agent->user_destroy = destroy;
+
+	dbus_pending_call_set_notify(agent->call, display_action_cb,
+					agent, NULL);
+
+	return 0;
+}
+
+static void confirm_open_channel_cb(DBusPendingCall *call, void *data)
+{
+	struct stk_agent *agent = data;
+	stk_agent_confirmation_cb cb = agent->user_cb;
+	DBusMessage *reply = dbus_pending_call_steal_reply(call);
+	enum stk_agent_result result;
+	gboolean remove_agent;
+	dbus_bool_t confirm;
+
+	if (check_error(agent, reply,
+			ALLOWED_ERROR_TERMINATE, &result) == -EINVAL) {
+		remove_agent = true;
+		goto error;
+	}
+
+	if (result != STK_AGENT_RESULT_OK) {
+		cb(result, FALSE, agent->user_data);
+		goto done;
+	}
+
+	if (dbus_message_get_args(reply, NULL,
+					DBUS_TYPE_BOOLEAN, &confirm,
+					DBUS_TYPE_INVALID) == FALSE) {
+		ofono_error("Can't parse the reply to ConfirmOpenChannel()");
+		remove_agent = true;
+		goto error;
+	}
+
+	cb(result, confirm, agent->user_data);
+
+	CALLBACK_END();
+}
+
+int stk_agent_confirm_open_channel(struct stk_agent *agent, const char *text,
+					const struct stk_icon_id *icon,
+					stk_agent_confirmation_cb cb,
+					void *user_data,
+					ofono_destroy_func destroy, int timeout)
+{
+	DBusConnection *conn = ofono_dbus_get_connection();
+
+	agent->msg = dbus_message_new_method_call(agent->bus, agent->path,
+							OFONO_SIM_APP_INTERFACE,
+							"ConfirmOpenChannel");
+	if (agent->msg == NULL)
+		return -ENOMEM;
+
+	dbus_message_append_args(agent->msg,
+					DBUS_TYPE_STRING, &text,
+					DBUS_TYPE_BYTE, &icon->id,
+					DBUS_TYPE_INVALID);
+
+	if (dbus_connection_send_with_reply(conn, agent->msg, &agent->call,
+						timeout) == FALSE ||
+						agent->call == NULL)
+		return -EIO;
+
+	agent->user_cb = cb;
+	agent->user_data = user_data;
+	agent->user_destroy = destroy;
+
+	dbus_pending_call_set_notify(agent->call, confirm_open_channel_cb,
+					agent, NULL);
+
+	return 0;
+}
